@@ -12,20 +12,23 @@ from __future__ import annotations
 
 import argparse
 import os
+import string
 from contextlib import asynccontextmanager
 from functools import partial
-from typing import Any, Dict, Optional
+from typing import Any, Dict, List, Optional
 
 import anyio
-from pydantic import Field, StrictInt
+from pydantic import Field
 from typing_extensions import Annotated
 
 # NOTE: For typing, import from the fastmcp package; fallback handled at runtime.
 try:  # Prefer the standalone fastmcp package when available
     from fastmcp import Context, FastMCP
+    from fastmcp.exceptions import NotFoundError
     from fastmcp.utilities.logging import configure_logging, get_logger
 except ImportError:  # Fall back to the implementation bundled with python-sdk
     from mcp.server.fastmcp import Context, FastMCP  # type: ignore[import-untyped,assignment]
+    from mcp.server.fastmcp.exceptions import NotFoundError  # type: ignore[import-untyped,assignment]
     from mcp.server.fastmcp.utilities.logging import configure_logging, get_logger  # type: ignore[import-untyped,assignment]
 
 from mcp_server_snowflake.server import (  # type: ignore[import-untyped]
@@ -41,6 +44,7 @@ from mcp_server_snowflake.utils import (  # type: ignore[import-untyped]
 
 from .config import Config, ConfigError, apply_config_overrides, get_config, load_config
 from .context import create_service_context
+from .logging import QueryHistory, update_cache_manifest_insight
 
 # Lineage functionality removed - not part of igloo-mcp
 from .mcp.tools import (  # QueryLineageTool,  # Removed - lineage functionality not part of igloo-mcp
@@ -51,12 +55,14 @@ from .mcp.tools import (  # QueryLineageTool,  # Removed - lineage functionality
     GetCatalogSummaryTool,
     HealthCheckTool,
     PreviewTableTool,
+    SearchCatalogTool,
 )
 from .mcp.utils import get_profile_recommendations
 from .mcp_health import (
     MCPHealthMonitor,
 )
 from .mcp_resources import MCPResourceManager
+from .path_utils import resolve_artifact_root
 from .profile_utils import (
     ProfileValidationError,
     get_profile_summary,
@@ -82,6 +88,27 @@ _resource_manager: Optional[MCPResourceManager] = None
 _catalog_service: Optional[CatalogService] = None
 
 
+def read_sql_artifact_by_sha(sql_sha256: str) -> str:
+    """Return the SQL text for the given SHA-256 hash."""
+
+    if len(sql_sha256) != 64 or any(ch not in string.hexdigits for ch in sql_sha256):
+        raise ValueError("sql_sha256 must be a 64-character hexadecimal digest")
+
+    artifact_root = resolve_artifact_root().resolve()
+    artifact_path = (
+        artifact_root / "queries" / "by_sha" / f"{sql_sha256}.sql"
+    ).resolve()
+    if not artifact_path.is_relative_to(artifact_root):
+        raise FileNotFoundError(
+            f"SQL artifact for {sql_sha256} not found under {artifact_root}"
+        )
+    if not artifact_path.exists() or not artifact_path.is_file():
+        raise FileNotFoundError(
+            f"SQL artifact for {sql_sha256} not found under {artifact_root}"
+        )
+    return artifact_path.read_text(encoding="utf-8")
+
+
 def _get_catalog_summary_sync(catalog_dir: str) -> Dict[str, Any]:
     service = _catalog_service
     if service is None:
@@ -158,6 +185,7 @@ def register_igloo_mcp(
     test_connection_inst = ConnectionTestTool(config, snowflake_service)
     health_check_inst = HealthCheckTool(config, snowflake_service, _health_monitor)
     get_catalog_summary_inst = GetCatalogSummaryTool(catalog_service)
+    search_catalog_inst = SearchCatalogTool()
 
     @server.tool(
         name="execute_query", description="Execute a SQL query against Snowflake"
@@ -187,9 +215,12 @@ def register_igloo_mcp(
             ),
         ] = None,
         timeout_seconds: Annotated[
-            Optional[StrictInt],
+            Optional[int],
             Field(
-                description="Query timeout in seconds (default: 30s from config)",
+                description=(
+                    "Query timeout in seconds (default: 30s from config). "
+                    "Integer between 1 and 3600."
+                ),
                 ge=1,
                 le=3600,
                 default=None,
@@ -202,20 +233,117 @@ def register_igloo_mcp(
                 default=False,
             ),
         ] = False,
+        post_query_insight: Annotated[
+            Optional[Dict[str, Any] | str],
+            Field(
+                description=(
+                    "Optional insights or key findings from query results. Metadata-only; no extra compute. "
+                    "Can be a summary string or structured JSON with key metrics and business impact."
+                ),
+                default=None,
+            ),
+        ] = None,
         ctx: Context | None = None,
     ) -> Dict[str, Any]:
         """Execute a SQL query against Snowflake - delegates to ExecuteQueryTool."""
-        return await execute_query_inst.execute(
-            statement=statement,
-            warehouse=warehouse,
-            database=database,
-            schema=schema,
-            role=role,
-            reason=reason,
-            timeout_seconds=timeout_seconds,
-            verbose_errors=verbose_errors,
-            ctx=ctx,
-        )
+        try:
+            return await execute_query_inst.execute(
+                statement=statement,
+                warehouse=warehouse,
+                database=database,
+                schema=schema,
+                role=role,
+                reason=reason,
+                timeout_seconds=timeout_seconds,
+                verbose_errors=verbose_errors,
+                post_query_insight=post_query_insight,
+                ctx=ctx,
+            )
+        except ValueError as e:
+            # Enhanced structured error for validation issues
+            error_msg = str(e)
+            if "timeout_seconds must be an integer" in error_msg:
+                raise ValueError(
+                    {
+                        "code": "PARAM_TYPE",
+                        "message": "Invalid parameter type: timeout_seconds must be an integer",
+                        "data": {
+                            "path": ".timeout_seconds",
+                            "expected": "integer",
+                            "received_type": "string",
+                            "hint": "Use 480 (number) without quotes instead of '480'",
+                            "examples": [30, 60, 300, 480],
+                        },
+                    }
+                ) from e
+            elif "timeout_seconds must be between" in error_msg:
+                raise ValueError(
+                    {
+                        "code": "PARAM_RANGE",
+                        "message": "Invalid parameter value: timeout_seconds out of range",
+                        "data": {
+                            "path": ".timeout_seconds",
+                            "min": 1,
+                            "max": 3600,
+                            "received_sample": timeout_seconds,
+                            "hint": "Use a timeout between 1 and 3600 seconds",
+                        },
+                    }
+                ) from e
+            elif "SQL statement type" in error_msg:
+                # Pass through enhanced SQL validation errors
+                raise ValueError(error_msg) from e
+            else:
+                # Generic parameter error
+                raise ValueError(
+                    {
+                        "code": "PARAM_VALIDATION",
+                        "message": "Parameter validation failed",
+                        "data": {
+                            "error": error_msg,
+                            "hint": "Check parameter values and types",
+                        },
+                    }
+                ) from e
+        except TimeoutError as e:
+            # Structured timeout error
+            raise RuntimeError(
+                {
+                    "code": "QUERY_TIMEOUT",
+                    "message": f"Query timeout after {timeout_seconds or 30}s",
+                    "data": {
+                        "timeout_seconds": timeout_seconds or 30,
+                        "suggestions": [
+                            "Increase timeout: timeout_seconds=480",
+                            "Add WHERE/LIMIT clause to reduce data volume",
+                            "Use larger warehouse for complex queries",
+                            "Use verbose_errors=True for detailed hints",
+                        ],
+                    },
+                }
+            ) from e
+        except Exception as e:
+            # Structured execution error
+            error_msg = str(e)
+            if verbose_errors:
+                # Include detailed error when requested
+                raise RuntimeError(error_msg) from e
+            else:
+                # Compact, actionable error
+                raise RuntimeError(
+                    {
+                        "code": "QUERY_EXECUTION_FAILED",
+                        "message": f"Query execution failed: {error_msg[:150]}",
+                        "data": {
+                            "hint": "Use verbose_errors=true for detailed information",
+                            "suggestions": [
+                                "Check SQL syntax and table names",
+                                "Verify database/schema context",
+                                "Check permissions for the objects referenced",
+                            ],
+                        },
+                    }
+                ) from e
 
     @server.tool(name="preview_table", description="Preview table contents")
     async def preview_table_tool(
@@ -320,6 +448,207 @@ def register_igloo_mcp(
         """Get catalog summary - delegates to GetCatalogSummaryTool."""
         return await get_catalog_summary_inst.execute(catalog_dir=catalog_dir)
 
+    @server.tool(
+        name="search_catalog", description="Search locally built catalog artifacts"
+    )
+    async def search_catalog_tool(
+        catalog_dir: Annotated[
+            str,
+            Field(
+                description="Directory containing catalog artifacts (catalog.json or catalog.jsonl).",
+                default="./data_catalogue",
+            ),
+        ] = "./data_catalogue",
+        object_types: Annotated[
+            Optional[List[str]],
+            Field(description="Optional list of object types to include", default=None),
+        ] = None,
+        database: Annotated[
+            Optional[str],
+            Field(description="Filter results to a specific database", default=None),
+        ] = None,
+        schema: Annotated[
+            Optional[str],
+            Field(description="Filter results to a specific schema", default=None),
+        ] = None,
+        name_contains: Annotated[
+            Optional[str],
+            Field(
+                description="Substring search on object name (case-insensitive)",
+                default=None,
+            ),
+        ] = None,
+        column_contains: Annotated[
+            Optional[str],
+            Field(
+                description="Substring search on column name (case-insensitive)",
+                default=None,
+            ),
+        ] = None,
+        limit: Annotated[
+            int,
+            Field(
+                description="Maximum number of results to return",
+                ge=1,
+                le=500,
+                default=20,
+            ),
+        ] = 20,
+    ) -> Dict[str, Any]:
+        return await search_catalog_inst.execute(
+            catalog_dir=catalog_dir,
+            object_types=object_types,
+            database=database,
+            schema=schema,
+            name_contains=name_contains,
+            column_contains=column_contains,
+            limit=limit,
+        )
+
+    @server.tool(
+        name="record_query_insight",
+        description="Attach a post-hoc insight to a prior execute_query run",
+    )
+    async def record_query_insight_tool(
+        execution_id: Annotated[
+            str,
+            Field(
+                description="Execution ID from execute_query.audit_info.execution_id"
+            ),
+        ],
+        post_query_insight: Annotated[
+            Dict[str, Any] | str,
+            Field(description="LLM-provided post-query insight"),
+        ],
+        cache_manifest_override: Annotated[
+            Optional[str],
+            Field(
+                description="Optional path to manifest if audit_info missing",
+                default=None,
+            ),
+        ] = None,
+        source: Annotated[
+            Optional[str],
+            Field(
+                description="Optional source identifier (e.g., 'human', 'agent:claude')",
+                default=None,
+            ),
+        ] = None,
+    ) -> Dict[str, Any]:
+        """Record a post-hoc insight for a prior query execution."""
+        history = QueryHistory.from_env()
+        result = history.record_insight(
+            execution_id=execution_id,
+            post_query_insight=post_query_insight,
+            source=source,
+        )
+
+        # Try to update manifest if path provided
+        updated_manifest: Optional[str] = None
+        if cache_manifest_override:
+            from pathlib import Path
+
+            manifest_path = Path(cache_manifest_override)
+            if update_cache_manifest_insight(manifest_path, post_query_insight):
+                updated_manifest = str(manifest_path.resolve())
+
+        return {
+            "execution_id": result["execution_id"],
+            "deduped": result["deduped"],
+            "updated_manifest": updated_manifest,
+        }
+
+    @server.tool(
+        name="describe_logging_paths",
+        description="Get resolved paths for history, artifacts, and cache with writability checks",
+    )
+    async def describe_logging_paths_tool() -> Dict[str, Any]:
+        """Describe logging paths and writability."""
+        from pathlib import Path
+
+        history = QueryHistory.from_env()
+        history_path = str(history.path) if history.path else None
+
+        artifact_root_raw = os.environ.get("IGLOO_MCP_ARTIFACT_ROOT")
+        from .path_utils import resolve_artifact_root
+
+        try:
+            artifact_root = resolve_artifact_root(raw=artifact_root_raw)
+            artifact_root_str = str(artifact_root)
+        except Exception:
+            artifact_root_str = None
+
+        cache_root_str = None
+        if artifact_root_str:
+            from .path_utils import DEFAULT_CACHE_SUBDIR
+
+            cache_root = Path(artifact_root_str) / DEFAULT_CACHE_SUBDIR
+            cache_root_str = str(cache_root.resolve())
+
+        # Check writability
+        writable: Dict[str, bool] = {}
+        if history_path:
+            try:
+                path = Path(history_path)
+                path.parent.mkdir(parents=True, exist_ok=True)
+                test_file = path.parent / ".writable_test"
+                test_file.write_text("test")
+                test_file.unlink()
+                writable["history"] = True
+            except Exception:
+                writable["history"] = False
+        else:
+            writable["history"] = False
+
+        if artifact_root_str:
+            try:
+                path = Path(artifact_root_str)
+                path.mkdir(parents=True, exist_ok=True)
+                test_file = path / ".writable_test"
+                test_file.write_text("test")
+                test_file.unlink()
+                writable["artifacts"] = True
+            except Exception:
+                writable["artifacts"] = False
+        else:
+            writable["artifacts"] = False
+
+        if cache_root_str:
+            try:
+                path = Path(cache_root_str)
+                path.mkdir(parents=True, exist_ok=True)
+                test_file = path / ".writable_test"
+                test_file.write_text("test")
+                test_file.unlink()
+                writable["cache"] = True
+            except Exception:
+                writable["cache"] = False
+        else:
+            writable["cache"] = False
+
+        return {
+            "history_path": history_path,
+            "artifact_root": artifact_root_str,
+            "cache_root": cache_root_str,
+            "writable": writable,
+        }
+
+    @server.resource(
+        "igloo://queries/by-sha/{sql_sha256}.sql",
+        name="sql_artifact_by_sha",
+        description="Full SQL text for a recorded query execution identified by its SHA-256 hash.",
+        mime_type="text/sql; charset=utf-8",
+    )
+    async def sql_artifact_by_sha(sql_sha256: str) -> str:
+        try:
+            return read_sql_artifact_by_sha(sql_sha256)
+        except FileNotFoundError as exc:
+            raise NotFoundError(str(exc)) from exc
+        except Exception as exc:  # pragma: no cover - unlikely I/O error
+            raise NotFoundError(
+                f"SQL artifact for {sql_sha256} is unreadable: {exc}"
+            ) from exc
+
     if enable_cli_bridge and snow_cli is not None:
 
         @server.tool(
